# Average fields along the field directions (i.e., Fw along the w-direction).  See the
# description of matrix/mean.jl/create_m̂().  Technically, apply_m̂!() can be used to average
# fields normal to the direction of averaging, but it is not used that way in apply_mean!().

# Assumes the space dimension and field dimension are the same.  In other words, when the
# space coordinate indices are (i,j,k), then the field has three vector components.
# Therefore, for the input field array F[i,j,k,w], we assume w = 1:3.

# The functions calculate the average and add to the output array, instead of replacing the
# values stored in the output array.  Therefore, if the derivative values themselves are
# desired, pass the output array initialized with zeros.

# Inside @threads, avoid defining β, which requires another let block to avoid unnecessary
# allocations.

export apply_m̂!, apply_mean!

# Wrapper for converting AbstractVector's to SVec's for arithmetic averaging
# This corresponds to the wrappers to create discrete versions of differential operators,
# but the nonzero enries in the matrix are 0.5 rather than 1.0.
apply_mean!(G::AbsArrNumber,  # output field; G[i,j,k,w] is w-component of G at (i,j,k)
            F::AbsArrNumber,  # input field; G[i,j,k,w] is w-component of G at (i,j,k)
            isfwd::AbsVecBool,  # isfwd[w] = true|false: m̂_w is forward|backward averaging
            isbloch::AbsVecBool=fill(true,length(isfwd)),  # isbloch[w] = true|false : w-boundary condition is Bloch|symmetric
            e⁻ⁱᵏᴸ::AbsVecNumber=ones(length(isfwd));  # e⁻ⁱᵏᴸ[w]: Bloch phase factor in w-direction
            α::Number=1.0,  # scale factor to multiply to result before adding it to G: G += α mean(F)
            overwrite::Bool=true) =  # false to add result to output storage G
    (K = length(isfwd); apply_mean!(G, F, SBool{K}(isfwd), SBool{K}(isbloch), SVec{K}(e⁻ⁱᵏᴸ); α, overwrite))

# Wrapper for converting AbstractVector's to SVec's for weighted averaging
apply_mean!(G::AbsArrNumber,  # output field; G[i,j,k,w] is w-component of G at (i,j,k)
            F::AbsArrNumber,  # input field; G[i,j,k,w] is w-component of G at (i,j,k)
            isfwd::AbsVecBool,  # isfwd[w] = true|false: m̂_w is forward|backward averaging
            ∆l::NTuple{K,AbsVecNumber},  # line segments to multiply with; vectors of length N
            ∆l′⁻¹::NTuple{K,AbsVecNumber},  # inverse of line segments to divide by; vectors of length N
            isbloch::AbsVecBool=fill(true,K),  # isbloch[w] = true|false : w-boundary condition is Bloch|symmetric
            e⁻ⁱᵏᴸ::AbsVecNumber=ones(K);  # e⁻ⁱᵏᴸ[w]: Bloch phase factor in w-direction
            α::Number=1.0,  # scale factor to multiply to result before adding it to G: G += α mean(F)
            overwrite::Bool=true  # false to add result to output storage G
            ) where {K} =
    apply_mean!(G, F, SBool{K}(isfwd), ∆l, ∆l′⁻¹, SBool{K}(isbloch), SVec{K}(e⁻ⁱᵏᴸ); α, overwrite)

# Concrete implementation for arithmetic averaging
# Initially this was implemented by passing a tuple of ones() os ∆l and ∆l′⁻¹, but because
# arithmetic averaging is used frequently, the optimized version is implemented separately.
function apply_mean!(G::AbsArrNumber{K₊₁},  # output field; G[i,j,k,w] is w-component of G at (i,j,k)
                     F::AbsArrNumber{K₊₁},  # input field; G[i,j,k,w] is w-component of G at (i,j,k)
                     isfwd::SBool{K},  # isfwd[w] = true|false: m̂_w is forward|backward averaging
                     isbloch::SBool{K},  # isbloch[w] = true|false : w-boundary condition is Bloch|symmetric
                     e⁻ⁱᵏᴸ::SNumber{K};  # e⁻ⁱᵏᴸ[w]: Bloch phase factor in w-direction
                     α::Number=1.0,  # scale factor to multiply to result before adding it to G: G += α mean(F)
                     overwrite::Bool=true  # false to add result to output storage G
                     ) where {K,K₊₁}  #  K is space dimension; K₊₁ = K + 1
    @assert K₊₁==K+1
    for nw = 1:K  # direction of averaging
        nv = nw  # component of output field generated by w-directional averaging operator m̂_w (row index of matrix block)
        Gᵥ = selectdim(G, K₊₁, nv)  # nv-th component of output field

        nu = nw  # component of input field to feed to w-directional averaging operator m̂_w (column index of matrix block)
        Fᵤ = selectdim(F, K₊₁, nu)  # nu-th component of input field

        apply_m̂!(Gᵥ, Fᵤ, nw, isfwd[nw], isbloch[nw], e⁻ⁱᵏᴸ[nw]; α, overwrite)
    end
end

# Concrete implementation for weighted averaging
# For the implementation details, see the comments in matrix/mean.jl.
function apply_mean!(G::AbsArrNumber{K₊₁},  # output field; G[i,j,k,w] is w-component of G at (i,j,k)
                     F::AbsArrNumber{K₊₁},  # input field; G[i,j,k,w] is w-component of G at (i,j,k)
                     isfwd::SBool{K},  # isfwd[w] = true|false: m̂_w is forward|backward averaging
                     ∆l::NTuple{K,AbsVecNumber},  # line segments to multiply with; vectors of length N
                     ∆l′⁻¹::NTuple{K,AbsVecNumber},  # inverse of line segments to divide by; vectors of length N
                     isbloch::SBool{K},  # isbloch[w] = true|false : w-boundary condition is Bloch|symmetric
                     e⁻ⁱᵏᴸ::SNumber{K};  # e⁻ⁱᵏᴸ[w]: Bloch phase factor in w-direction
                     α::Number=1.0,  # scale factor to multiply to result before adding it to G: G += α mean(F)
                     overwrite::Bool=true  # false to add result to output storage G
                     ) where {K,K₊₁}  #  K is space dimension; K₊₁ = K + 1
    @assert K₊₁==K+1
    for nw = 1:K  # direction of averaging
        nv = nw  # component of output field generated by w-directional averaging operator m̂_w (row index of matrix block)
        Gᵥ = selectdim(G, K₊₁, nv)  # nv-th component of output field

        nu = nw  # component of input field to feed to w-directional averaging operator m̂_w (column index of matrix block)
        Fᵤ = selectdim(F, K₊₁, nu)  # nu-th component of input field

        apply_m̂!(Gᵥ, Fᵤ, nw, isfwd[nw], ∆l[nw], ∆l′⁻¹[nw], isbloch[nw], e⁻ⁱᵏᴸ[nw]; α, overwrite)
    end
end

## Field-averaging operators "m̂" (as used in de Moerloose and de Zutter)
#
# This applies the averaging operator for a single Cartesian component.  For the operator
# for all three Cartesian components, use apply_mean!.

# Below, I set the default value for e⁻ⁱᵏᴸ below, even though I don't usually set default
# values for the arguments in concrete implementation (except for keyword arguments, which
# require default values).  This is to make the arguments of apply_m̂!() to have default
# values when the corresponding arguments of create_m̂() do.

# Arithmetic m̂ for 3D
function apply_m̂!(Gᵥ::AbsArrNumber{3},  # v-component of output field (v = x, y, z)
                  Fᵤ::AbsArrNumber{3},  # u-component of input field (u = x, y, z)
                  nw::Integer,  # 1|2|3 for averaging along x|y|z
                  isfwd::Bool,  # true|false for forward|backward averaging
                  isbloch::Bool,  # boundary condition in w-direction
                  e⁻ⁱᵏᴸ::Number=1.0;  # Bloch phase factor
                  α::Number=1.0,  # scale factor to multiply to result before adding it to Gᵥ: Gᵥ += α m(Fᵤ)
                  overwrite::Bool=true)  # false to add result to output storage G
    @assert size(Gᵥ)==size(Fᵤ)

    Nx, Ny, Nz = size(Fᵤ)
    α2 = 0.5α

    # Make sure not to include branches inside for loops.
    overwrite && (Gᵥ .= 0)
    if isfwd
        if nw == 1  # w = x
            if isbloch
                # 1. At locations except for the positive end of the x-direction
                @threads for k = 1:Nz
                    for j = 1:Ny, i = 1:Nx-1
                        @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i+1,j,k] + Fᵤ[i,j,k])
                    end
                end

                # 2. At the positive end of the x-direction (where the boundary fields are
                # taken from the negative-end boundary)
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[Nx,j,k] += α2 * (e⁻ⁱᵏᴸ*Fᵤ[1,j,k] + Fᵤ[Nx,j,k])  # Fᵤ[Nx+1,j,k] = exp(-i kx Lx) * Fᵤ[1,j,k]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # x-direction
                @threads for k = 1:Nz
                    for j = 1:Ny, i = 2:Nx-1
                        @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i+1,j,k] + Fᵤ[i,j,k])
                    end
                end

                # 2. At the negative end of the x-direction (where the boundary fields are
                # assumed zero)
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[1,j,k] += α2 * Fᵤ[2,j,k]  # Fᵤ[1,j,k] == 0
                end

                # 3. At the positive end of the x-direction (where the boundary fields are
                # assumed zero)
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[Nx,j,k] += α2 * Fᵤ[Nx,j,k]  # Fᵤ[Nx+1,j,k] == 0
                end
            end
        elseif nw == 2  # w = y
            if isbloch
                # 1. At locations except for the positive end of the y-direction
                @threads for k = 1:Nz
                    for j = 1:Ny-1, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i,j+1,k] + Fᵤ[i,j,k])
                    end
                end

                # 2. At the positive end of the y-direction (where the boundary fields are
                # taken from the negative-end boundary)
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,Ny,k] += α2 * (e⁻ⁱᵏᴸ*Fᵤ[i,1,k] + Fᵤ[i,Ny,k])  # Fᵤ[i,Ny+1,k] = exp(-i ky Ly) * Fᵤ[i,1,k]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # y-direction
                @threads for k = 1:Nz
                    for j = 2:Ny-1, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i,j+1,k] + Fᵤ[i,j,k])
                    end
                end

                # 2. At the negative end of the y-direction (where the boundary fields are
                # assumed zero)
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,1,k] += α2 * Fᵤ[i,2,k]  # Fᵤ[i,1,k] == 0
                end

                # 3. At the positive end of the y-direction (where the boundary fields are
                # assumed zero)
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,Ny,k] += α2 * Fᵤ[i,Ny,k]  # Fᵤ[i,Ny+1,k] == 0
                end
            end
        else  # nw == 3; w = z
            if isbloch
                # 1. At locations except for the positive end of the z-direction
                @threads for k = 1:Nz-1
                    for j = 1:Ny, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i,j,k+1] + Fᵤ[i,j,k])
                    end
                end

                # 2. At the positive end of the z-direction (where the boundary fields are
                # taken from the negative-end boundary)
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,Nz] += α2 * (e⁻ⁱᵏᴸ*Fᵤ[i,j,1] + Fᵤ[i,j,Nz])  # Fᵤ[i,j,Nz+1] = exp(-i kz Lz) * Fᵤ[i,j,1]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # z-direction
                @threads for k = 2:Nz-1
                    for j = 1:Ny, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i,j,k+1] + Fᵤ[i,j,k])
                    end
                end

                # 2. At the negative end of the z-direction (where the boundary fields are
                # assumed zero)
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,1] += α2 * Fᵤ[i,j,2]  # Fᵤ[i,j,1] == 0
                end

                # 3. At the positive end of the z-direction (where the boundary fields are
                # assumed zero)
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,Nz] += α2 * Fᵤ[i,j,Nz]  # Fᵤ[i,j,Nz+1] == 0
                end
            end
        end  # if nw == ...
    else  # backward averaging
        if nw == 1  # w = x
            # 1. At the locations except for the negative end of the x-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for k = 1:Nz
                for j = 1:Ny, i = 2:Nx  # not i = 2:Nx-1
                    @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i,j,k] + Fᵤ[i-1,j,k])
                end
            end

            # 2. At the negative end of the x-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[1,j,k] += α2 * (Fᵤ[1,j,k] + Fᵤ[Nx,j,k]/e⁻ⁱᵏᴸ)  # Fᵤ[0,j,k] = Fᵤ[Nx,j,k] / exp(-i kx Lx)
                end
            else  # symmetry boundary
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[1,j,k] += α * Fᵤ[1,j,k]  # Fᵤ[0,j,k] = Fᵤ[1,j,k]; use α instead of α2 as field is doubled
                end
            end
        elseif nw == 2  # w = y
            # 1. At the locations except for the negative end of the y-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for k = 1:Nz
                for j = 2:Ny, i = 1:Nx  # not j = 2:Ny-1
                    @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i,j,k] + Fᵤ[i,j-1,k])
                end
            end

            # 2. At the negative end of the y-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,1,k] += α2 * (Fᵤ[i,1,k] + Fᵤ[i,Ny,k]/e⁻ⁱᵏᴸ)  # Fᵤ[i,0,k] = Fᵤ[0,Ny,k] / exp(-i ky Ly)
                end
            else  # symmetry boundary
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,1,k] += α * Fᵤ[i,1,k]  # Fᵤ[i,0,k] = Fᵤ[i,1,k]; use α instead of α2 as field is doubled
                end
            end
        else  # nw == 3; w = z
            # 1. At the locations except for the negative end of the z-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for k = 2:Nz
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,k] += α2 * (Fᵤ[i,j,k] + Fᵤ[i,j,k-1])
                end
            end

            # 2. At the negative end of the z-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,1] += α2 * (Fᵤ[i,j,1] + Fᵤ[i,j,Nz]/e⁻ⁱᵏᴸ)  # Fᵤ[i,j,0] = Fᵤ[i,j,Nz] / exp(-i kz Lz)
                end
            else  # symmetry boundary
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,1] += α * Fᵤ[i,j,1]  # Fᵤ[i,j,0] = Fᵤ[i,j,Nz]; use α instead of α2 as field is doubled
                end
            end
        end  # if nw == ...
    end  # if isfwd

    return nothing
end

# Arithmetic m̂ for 2D
function apply_m̂!(Gᵥ::AbsArrNumber{2},  # v-component of output field (v = x, y)
                  Fᵤ::AbsArrNumber{2},  # u-component of input field (u = x, y)
                  nw::Integer,  # 1|2 for averaging along x|y
                  isfwd::Bool,  # true|false for forward|backward averaging
                  isbloch::Bool,  # boundary condition in w-direction
                  e⁻ⁱᵏᴸ::Number=1.0;  # Bloch phase factor
                  α::Number=1.0,  # scale factor to multiply to result before adding it to Gᵥ: Gᵥ += α m(Fᵤ)
                  overwrite::Bool=true)  # false to add result to output storage G
    @assert size(Gᵥ)==size(Fᵤ)

    Nx, Ny = size(Fᵤ)
    α2 = 0.5α

    # Make sure not to include branches inside for loops.
    overwrite && (Gᵥ .= 0)
    if isfwd
        if nw == 1  # w = x
            if isbloch
                # 1. At locations except for the positive end of the x-direction
                @threads for j = 1:Ny
                    for i = 1:Nx-1
                        @inbounds Gᵥ[i,j] += α2 * (Fᵤ[i+1,j] + Fᵤ[i,j])
                    end
                end

                # 2. At the positive end of the x-direction (where the boundary fields are
                # taken from the negative-end boundary)
                for j = 1:Ny
                    @inbounds Gᵥ[Nx,j] += α2 * (e⁻ⁱᵏᴸ*Fᵤ[1,j] + Fᵤ[Nx,j])  # Fᵤ[Nx+1,j] = exp(-i kx Lx) * Fᵤ[1,j]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # x-direction
                @threads for j = 1:Ny
                    for i = 2:Nx-1
                        @inbounds Gᵥ[i,j] += α2 * (Fᵤ[i+1,j] + Fᵤ[i,j])
                    end
                end

                # 2. At the negative end of the x-direction (where the boundary fields are
                # assumed zero)
                for j = 1:Ny
                    @inbounds Gᵥ[1,j] += α2 * Fᵤ[2,j]  # Fᵤ[1,j] == 0
                end

                # 3. At the positive end of the x-direction (where the boundary fields are
                # assumed zero)
                for j = 1:Ny
                    @inbounds Gᵥ[Nx,j] += α2 * Fᵤ[Nx,j]  # Fᵤ[Nx+1,j] == 0
                end
            end
        else  # nw == 2; w = y
            if isbloch
                # 1. At locations except for the positive end of the y-direction
                @threads for j = 1:Ny-1
                    for i = 1:Nx
                        @inbounds Gᵥ[i,j] += α2 * (Fᵤ[i,j+1] + Fᵤ[i,j])
                    end
                end

                # 2. At the positive end of the y-direction (where the boundary fields are
                # taken from the negative-end boundary)
                for i = 1:Nx
                    @inbounds Gᵥ[i,Ny] += α2 * (e⁻ⁱᵏᴸ*Fᵤ[i,1] + Fᵤ[i,Ny])  # Fᵤ[i,Ny+1] = exp(-i ky Ly) * Fᵤ[i,1]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # y-direction
                @threads for j = 2:Ny-1
                    for i = 1:Nx
                        @inbounds Gᵥ[i,j] += α2 * (Fᵤ[i,j+1] + Fᵤ[i,j])
                    end
                end

                # 2. At the negative end of the y-direction (where the boundary fields are
                # assumed zero)
                for i = 1:Nx
                    @inbounds Gᵥ[i,1] += α2 * Fᵤ[i,2]  # Fᵤ[i,1] == 0
                end

                # 3. At the positive end of the y-direction (where the boundary fields are
                # assumed zero)
                for i = 1:Nx
                    @inbounds Gᵥ[i,Ny] += α2 * Fᵤ[i,Ny]  # Fᵤ[i,Ny+1] == 0
                end
            end
        end  # if nw == ...
    else  # backward averaging
        if nw == 1  # w = x
            # 1. At the locations except for the negative end of the x-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for j = 1:Ny
                for i = 2:Nx  # not i = 2:Nx-1
                    @inbounds Gᵥ[i,j] += α2 * (Fᵤ[i,j] + Fᵤ[i-1,j])
                end
            end

            # 2. At the negative end of the x-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                for j = 1:Ny
                    @inbounds Gᵥ[1,j] += α2 * (Fᵤ[1,j] + Fᵤ[Nx,j]/e⁻ⁱᵏᴸ)  # Fᵤ[0,j] = Fᵤ[Nx,j] / exp(-i kx Lx)
                end
            else  # symmetry boundary
                for j = 1:Ny
                    @inbounds Gᵥ[1,j] += α * Fᵤ[1,j]  # Fᵤ[0,j] = Fᵤ[1,j]; use α instead of α2 as field is doubled
                end
            end
        elseif nw == 2  # w = y
            # 1. At the locations except for the negative end of the y-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for j = 2:Ny
                for i = 1:Nx
                    @inbounds Gᵥ[i,j] += α2 * (Fᵤ[i,j] + Fᵤ[i,j-1])
                end
            end

            # 2. At the negative end of the y-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                for i = 1:Nx
                    @inbounds Gᵥ[i,1] += α2 * (Fᵤ[i,1] + Fᵤ[i,Ny]/e⁻ⁱᵏᴸ)  # Fᵤ[i,0] = Fᵤ[0,Ny] / exp(-i ky Ly)
                end
            else  # symmetry boundary
                for i = 1:Nx
                    @inbounds Gᵥ[i,1] += α * Fᵤ[i,1]  # Fᵤ[i,0] = Fᵤ[i,1]; use α instead of α2 as field is doubled
                end
            end
        end  # if nw == ...
    end  # if isfwd

    return nothing
end

# Arithmetic m̂ for 1D
function apply_m̂!(Gᵥ::AbsArrNumber{1},  # v-component of output field (v = x, y)
                  Fᵤ::AbsArrNumber{1},  # u-component of input field (u = x, y)
                  nw::Integer,  # 1|2 for averaging along x|y
                  isfwd::Bool,  # true|false for forward|backward averaging
                  isbloch::Bool,  # boundary condition in w-direction
                  e⁻ⁱᵏᴸ::Number=1.0;  # Bloch phase factor
                  α::Number=1.0,  # scale factor to multiply to result before adding it to Gᵥ: Gᵥ += α m(Fᵤ)
                  overwrite::Bool=true)  # false to add result to output storage G
    @assert size(Gᵥ)==size(Fᵤ)

    Nx = length(Fᵤ)  # not size(Fᵤ) unlike code for 2D and 3D
    α2 = 0.5α

    # Make sure not to include branches inside for loops.
    overwrite && (Gᵥ .= 0)
    if isfwd
        if isbloch
            # 1. At locations except for the positive end of the x-direction
            @threads for i = 1:Nx-1
                @inbounds Gᵥ[i] += α2 * (Fᵤ[i+1] + Fᵤ[i])
            end

            # 2. At the positive end of the x-direction (where the boundary fields are taken
            # from the negative-end boundary)
            @inbounds Gᵥ[Nx] += α2 * (e⁻ⁱᵏᴸ*Fᵤ[1] + Fᵤ[Nx])  # Fᵤ[Nx+1] = exp(-i kx Lx) * Fᵤ[1]
        else  # symmetry boundary
            # 1. At the locations except for the positive and negative ends of the x-direction
            @threads for i = 2:Nx-1
                @inbounds Gᵥ[i] += α2 * (Fᵤ[i+1] + Fᵤ[i])
            end

            # 2. At the negative end of the x-direction (where the boundary fields are assumed
            # zero)
            @inbounds Gᵥ[1] += α2 * Fᵤ[2]  # Fᵤ[1] == 0

            # 3. At the positive end of the x-direction (where the boundary fields are assumed
            # zero)
            @inbounds Gᵥ[Nx] += α2 * Fᵤ[Nx]  # Fᵤ[Nx+1] == 0
        end
    else  # backward averaging
        # 1. At the locations except for the negative end of the x-direction; unlike for the
        # forward difference, for the backward difference this part of the code is common
        # for both the Bloch and symmetry boundary conditions.
        @threads for i = 2:Nx
            @inbounds Gᵥ[i] += α2 * (Fᵤ[i] + Fᵤ[i-1])
        end

        # 2. At the negative end of the x-direction (where the boundary fields are taken
        # from the positive-end boundary for the Bloch boundary condition)
        if isbloch
            @inbounds Gᵥ[1] += α2 * (Fᵤ[1] + Fᵤ[Nx]/e⁻ⁱᵏᴸ)  # Fᵤ[0] = Fᵤ[Nx] / exp(-i kx Lx)
        else  # symmetry boundary
            @inbounds Gᵥ[1] += α * Fᵤ[1]  # Fᵤ[0] = Fᵤ[1]; use α instead of α2 as field is doubled
        end
    end  # if isfwd

    return nothing
end


# Weighted m̂ for 3D
function apply_m̂!(Gᵥ::AbsArrNumber{3},  # v-component of output field (v = x, y, z)
                  Fᵤ::AbsArrNumber{3},  # u-component of input field (u = x, y, z)
                  nw::Integer,  # 1|2|3 for averaging along x|y|z
                  isfwd::Bool,  # true|false for forward|backward averaging
                  ∆w::AbsVecNumber,  # line segments to multiply with; vector of length N[nw]
                  ∆w′⁻¹::AbsVecNumber,  # inverse of line segments to divide by; vector of length N[nw]
                  isbloch::Bool,  # boundary condition in w-direction
                  e⁻ⁱᵏᴸ::Number=1.0;  # Bloch phase factor
                  α::Number=1.0,  # scale factor to multiply to result before adding it to Gᵥ: Gᵥ += α m(Fᵤ)
                  overwrite::Bool=true)  # false to add result to output storage G
    @assert size(Gᵥ)==size(Fᵤ)
    @assert size(Fᵤ,nw)==length(∆w)
    @assert length(∆w)==length(∆w′⁻¹)

    Nx, Ny, Nz = size(Fᵤ)
    α2 = 0.5α

    # Make sure not to include branches inside for loops.
    overwrite && (Gᵥ .= 0)
    if isfwd
        if nw == 1  # w = x
            if isbloch
                # 1. At locations except for the positive end of the x-direction
                @threads for k = 1:Nz
                    for j = 1:Ny, i = 1:Nx-1
                        @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[i]) * (∆w[i+1]*Fᵤ[i+1,j,k] + ∆w[i]*Fᵤ[i,j,k])
                    end
                end

                # 2. At the positive end of the x-direction (where the boundary fields are
                # taken from the negative-end boundary)
                β = α2 * ∆w′⁻¹[Nx]
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[Nx,j,k] += β * (∆w[1]*e⁻ⁱᵏᴸ*Fᵤ[1,j,k] + ∆w[Nx]*Fᵤ[Nx,j,k])  # Fᵤ[Nx+1,j,k] = exp(-i kx Lx) * Fᵤ[1,j,k]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # x-direction
                @threads for k = 1:Nz
                    for j = 1:Ny, i = 2:Nx-1
                        @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[i]) * (∆w[i+1]*Fᵤ[i+1,j,k] + ∆w[i]*Fᵤ[i,j,k])
                    end
                end

                # 2. At the negative end of the x-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[1]
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[1,j,k] += β * ∆w[2]*Fᵤ[2,j,k]  # Fᵤ[1,j,k] == 0
                end

                # 3. At the positive end of the x-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[Nx]
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[Nx,j,k] += β * ∆w[Nx]*Fᵤ[Nx,j,k]  # Fᵤ[Nx+1,j,k] == 0
                end
            end
        elseif nw == 2  # w = y
            if isbloch
                # 1. At locations except for the positive end of the y-direction
                @threads for k = 1:Nz
                    for j = 1:Ny-1, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[j]) * (∆w[j+1]*Fᵤ[i,j+1,k] + ∆w[j]*Fᵤ[i,j,k])
                    end
                end

                # 2. At the positive end of the y-direction (where the boundary fields are
                # taken from the negative-end boundary)
                β = α2 * ∆w′⁻¹[Ny]
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,Ny,k] += β * (∆w[1]*e⁻ⁱᵏᴸ*Fᵤ[i,1,k] + ∆w[Ny]*Fᵤ[i,Ny,k])  # Fᵤ[i,Ny+1,k] = exp(-i ky Ly) * Fᵤ[i,1,k]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # y-direction
                @threads for k = 1:Nz
                    for j = 2:Ny-1, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[j]) * (∆w[j+1]*Fᵤ[i,j+1,k] + ∆w[j]*Fᵤ[i,j,k])
                    end
                end

                # 2. At the negative end of the y-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[1]
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,1,k] += β * ∆w[2]*Fᵤ[i,2,k]  # Fᵤ[i,1,k] == 0
                end

                # 3. At the positive end of the y-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[Ny]
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,Ny,k] += β * ∆w[Ny]*Fᵤ[i,Ny,k]  # Fᵤ[i,Ny+1,k] == 0
                end
            end
        else  # nw == 3; w = z
            if isbloch
                # 1. At locations except for the positive end of the z-direction
                @threads for k = 1:Nz-1
                    for j = 1:Ny, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[k]) * (∆w[k+1]*Fᵤ[i,j,k+1] + ∆w[k]*Fᵤ[i,j,k])
                    end
                end

                # 2. At the positive end of the z-direction (where the boundary fields are
                # taken from the negative-end boundary)
                β = α2 * ∆w′⁻¹[Nz]
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,Nz] += β * (∆w[1]*e⁻ⁱᵏᴸ*Fᵤ[i,j,1] + ∆w[Nz]*Fᵤ[i,j,Nz])  # Fᵤ[i,j,Nz+1] = exp(-i kz Lz) * Fᵤ[i,j,1]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # z-direction
                @threads for k = 2:Nz-1
                    for j = 1:Ny, i = 1:Nx
                        @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[k]) * (∆w[k+1]*Fᵤ[i,j,k+1] + ∆w[k]*Fᵤ[i,j,k])
                    end
                end

                # 2. At the negative end of the z-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[1]
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,1] += β * ∆w[2]*Fᵤ[i,j,2]  # Fᵤ[i,j,1] == 0
                end

                # 3. At the positive end of the z-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[Nz]
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,Nz] += β * ∆w[Nz]*Fᵤ[i,j,Nz]  # Fᵤ[i,j,Nz+1] == 0
                end
            end
        end  # if nw == ...
    else  # backward averaging
        if nw == 1  # w = x
            # 1. At the locations except for the negative end of the x-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for k = 1:Nz
                for j = 1:Ny, i = 2:Nx  # not i = 2:Nx-1
                    @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[i]) * (∆w[i]*Fᵤ[i,j,k] + ∆w[i-1]*Fᵤ[i-1,j,k])
                end
            end

            # 2. At the negative end of the x-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                β = α2 * ∆w′⁻¹[1]
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[1,j,k] += β * (∆w[1]*Fᵤ[1,j,k] + ∆w[Nx]*Fᵤ[Nx,j,k]/e⁻ⁱᵏᴸ)  # Fᵤ[0,j,k] = Fᵤ[Nx,j,k] / exp(-i kx Lx)
                end
            else  # symmetry boundary
                β = α * ∆w′⁻¹[1]
                for k = 1:Nz, j = 1:Ny
                    @inbounds Gᵥ[1,j,k] += β * ∆w[1]*Fᵤ[1,j,k]  # Fᵤ[0,j,k] = Fᵤ[1,j,k]
                end
            end
        elseif nw == 2  # w = y
            # 1. At the locations except for the negative end of the y-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for k = 1:Nz
                for j = 2:Ny, i = 1:Nx  # not j = 2:Ny-1
                    @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[j]) * (∆w[j]*Fᵤ[i,j,k] + ∆w[j-1]*Fᵤ[i,j-1,k])
                end
            end

            # 2. At the negative end of the y-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                β = α2 * ∆w′⁻¹[1]
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,1,k] += β * (∆w[1]*Fᵤ[i,1,k] + ∆w[Ny]*Fᵤ[i,Ny,k]/e⁻ⁱᵏᴸ)  # Fᵤ[i,0,k] = Fᵤ[0,Ny,k] / exp(-i ky Ly)
                end
            else  # symmetry boundary
                β = α * ∆w′⁻¹[1]
                for k = 1:Nz, i = 1:Nx
                    @inbounds Gᵥ[i,1,k] += β * ∆w[1]*Fᵤ[i,1,k]  # Fᵤ[i,0,k] = Fᵤ[i,1,k]
                end
            end
        else  # nw == 3; w = z
            # 1. At the locations except for the negative end of the z-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for k = 2:Nz
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,k] += (α2 * ∆w′⁻¹[k]) * (∆w[k]*Fᵤ[i,j,k] + ∆w[k-1]*Fᵤ[i,j,k-1])
                end
            end

            # 2. At the negative end of the z-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                β = α2 * ∆w′⁻¹[1]
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,1] += β * (∆w[1]*Fᵤ[i,j,1] + ∆w[Nz]*Fᵤ[i,j,Nz]/e⁻ⁱᵏᴸ)  # Fᵤ[i,j,0] = Fᵤ[i,j,Nz] / exp(-i kz Lz)
                end
            else  # symmetry boundary
                β = α * ∆w′⁻¹[1]
                for j = 1:Ny, i = 1:Nx
                    @inbounds Gᵥ[i,j,1] += β * ∆w[1]*Fᵤ[i,j,1]  # Fᵤ[i,j,0] = Fᵤ[i,j,Nz]
                end
            end
        end  # if nw == ...
    end  # if isfwd

    return nothing
end

# Weighted m̂ for 2D
function apply_m̂!(Gᵥ::AbsArrNumber{2},  # v-component of output field (v = x, y)
                  Fᵤ::AbsArrNumber{2},  # u-component of input field (u = x, y)
                  nw::Integer,  # 1|2 for averaging along x|y
                  isfwd::Bool,  # true|false for forward|backward averaging
                  ∆w::AbsVecNumber,  # line segments to multiply with; vector of length N[nw]
                  ∆w′⁻¹::AbsVecNumber,  # inverse of line segments to divide by; vector of length N[nw]
                  isbloch::Bool,  # boundary condition in w-direction
                  e⁻ⁱᵏᴸ::Number=1.0;  # Bloch phase factor
                  α::Number=1.0,  # scale factor to multiply to result before adding it to Gᵥ: Gᵥ += α m(Fᵤ)
                  overwrite::Bool=true)  # false to add result to output storage G
    @assert size(Gᵥ)==size(Fᵤ)
    @assert size(Fᵤ,nw)==length(∆w)
    @assert length(∆w)==length(∆w′⁻¹)

    Nx, Ny = size(Fᵤ)
    α2 = 0.5α

    # Make sure not to include branches inside for loops.
    overwrite && (Gᵥ .= 0)
    if isfwd
        if nw == 1  # w = x
            if isbloch
                # 1. At locations except for the positive end of the x-direction
                @threads for j = 1:Ny
                    for i = 1:Nx-1
                        @inbounds Gᵥ[i,j] += (α2 * ∆w′⁻¹[i]) * (∆w[i+1]*Fᵤ[i+1,j] + ∆w[i]*Fᵤ[i,j])
                    end
                end

                # 2. At the positive end of the x-direction (where the boundary fields are
                # taken from the negative-end boundary)
                β = α2 * ∆w′⁻¹[Nx]
                for j = 1:Ny
                    @inbounds Gᵥ[Nx,j] += β * (∆w[1]*e⁻ⁱᵏᴸ*Fᵤ[1,j] + ∆w[Nx]*Fᵤ[Nx,j])  # Fᵤ[Nx+1,j] = exp(-i kx Lx) * Fᵤ[1,j]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # x-direction
                @threads for j = 1:Ny
                    for i = 2:Nx-1
                        @inbounds Gᵥ[i,j] += (α2 * ∆w′⁻¹[i]) * (∆w[i+1]*Fᵤ[i+1,j] + ∆w[i]*Fᵤ[i,j])
                    end
                end

                # 2. At the negative end of the x-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[1]
                for j = 1:Ny
                    @inbounds Gᵥ[1,j] += β * ∆w[2]*Fᵤ[2,j]  # Fᵤ[1,j] == 0
                end

                # 3. At the positive end of the x-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[Nx]
                for j = 1:Ny
                    @inbounds Gᵥ[Nx,j] += β * ∆w[Nx]*Fᵤ[Nx,j]  # Fᵤ[Nx+1,j] == 0
                end
            end
        else  # nw == 2; w = y
            if isbloch
                # 1. At locations except for the positive end of the y-direction
                @threads for j = 1:Ny-1
                    for i = 1:Nx
                        @inbounds Gᵥ[i,j] += (α2 * ∆w′⁻¹[j]) * (∆w[j+1]*Fᵤ[i,j+1] + ∆w[j]*Fᵤ[i,j])
                    end
                end

                # 2. At the positive end of the y-direction (where the boundary fields are
                # taken from the negative-end boundary)
                β = α2 * ∆w′⁻¹[Ny]
                for i = 1:Nx
                    @inbounds Gᵥ[i,Ny] += β * (∆w[1]*e⁻ⁱᵏᴸ*Fᵤ[i,1] + ∆w[Ny]*Fᵤ[i,Ny])  # Fᵤ[i,Ny+1] = exp(-i ky Ly) * Fᵤ[i,1]
                end
            else  # symmetry boundary
                # 1. At the locations except for the positive and negative ends of the
                # y-direction
                @threads for j = 2:Ny-1
                    for i = 1:Nx
                        @inbounds Gᵥ[i,j] += (α2 * ∆w′⁻¹[j]) * (∆w[j+1]*Fᵤ[i,j+1] + ∆w[j]*Fᵤ[i,j])
                    end
                end

                # 2. At the negative end of the y-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[1]
                for i = 1:Nx
                    @inbounds Gᵥ[i,1] += β * ∆w[2]*Fᵤ[i,2]  # Fᵤ[i,1] == 0
                end

                # 3. At the positive end of the y-direction (where the boundary fields are
                # assumed zero)
                β = α2 * ∆w′⁻¹[Ny]
                for i = 1:Nx
                    @inbounds Gᵥ[i,Ny] += β * ∆w[Ny]*Fᵤ[i,Ny]  # Fᵤ[i,Ny+1] == 0
                end
            end
        end  # if nw == ...
    else  # backward averaging
        if nw == 1  # w = x
            # 1. At the locations except for the negative end of the x-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for j = 1:Ny
                for i = 2:Nx  # not i = 2:Nx-1
                    @inbounds Gᵥ[i,j] += (α2 * ∆w′⁻¹[i]) * (∆w[i]*Fᵤ[i,j] + ∆w[i-1]*Fᵤ[i-1,j])
                end
            end

            # 2. At the negative end of the x-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                β = α2 * ∆w′⁻¹[1]
                for j = 1:Ny
                    @inbounds Gᵥ[1,j] += β * (∆w[1]*Fᵤ[1,j] + ∆w[Nx]*Fᵤ[Nx,j]/e⁻ⁱᵏᴸ)  # Fᵤ[0,j] = Fᵤ[Nx,j] / exp(-i kx Lx)
                end
            else  # symmetry boundary
                β = α * ∆w′⁻¹[1]
                for j = 1:Ny
                    @inbounds Gᵥ[1,j] += β * ∆w[1]*Fᵤ[1,j]  # Fᵤ[0,j] = Fᵤ[1,j]
                end
            end
        elseif nw == 2  # w = y
            # 1. At the locations except for the negative end of the y-direction; unlike for
            # the forward difference, for the backward difference this part of the code is
            # common for both the Bloch and symmetry boundary conditions.
            @threads for j = 2:Ny
                for i = 1:Nx
                    @inbounds Gᵥ[i,j] += (α2 * ∆w′⁻¹[j]) * (∆w[j]*Fᵤ[i,j] + ∆w[j-1]*Fᵤ[i,j-1])
                end
            end

            # 2. At the negative end of the y-direction (where the boundary fields are taken
            # from the positive-end boundary for the Bloch boundary condition)
            if isbloch
                β = α2 * ∆w′⁻¹[1]
                for i = 1:Nx
                    @inbounds Gᵥ[i,1] += β * (∆w[1]*Fᵤ[i,1] + ∆w[Ny]*Fᵤ[i,Ny]/e⁻ⁱᵏᴸ)  # Fᵤ[i,0] = Fᵤ[0,Ny] / exp(-i ky Ly)
                end
            else  # symmetry boundary
                β = α * ∆w′⁻¹[1]
                for i = 1:Nx
                    @inbounds Gᵥ[i,1] += β * ∆w[1]*Fᵤ[i,1]  # Fᵤ[i,0] = Fᵤ[i,1]
                end
            end
        end  # if nw == ...
    end  # if isfwd

    return nothing
end

# Weighted m̂ for 1D
function apply_m̂!(Gᵥ::AbsArrNumber{1},  # v-component of output field (v = x, y)
                  Fᵤ::AbsArrNumber{1},  # u-component of input field (u = x, y)
                  nw::Integer,  # 1|2 for averaging along x|y
                  isfwd::Bool,  # true|false for forward|backward averaging
                  ∆w::AbsVecNumber,  # line segments to multiply with; vector of length N[nw]
                  ∆w′⁻¹::AbsVecNumber,  # inverse of line segments to divide by; vector of length N[nw]
                  isbloch::Bool,  # boundary condition in w-direction
                  e⁻ⁱᵏᴸ::Number=1.0;  # Bloch phase factor
                  α::Number=1.0,  # scale factor to multiply to result before adding it to Gᵥ: Gᵥ += α m(Fᵤ)
                  overwrite::Bool=true)  # false to add result to output storage G
    @assert size(Gᵥ)==size(Fᵤ)
    @assert size(Fᵤ,nw)==length(∆w)
    @assert length(∆w)==length(∆w′⁻¹)

    Nx = length(Fᵤ)  # not size(Fᵤ) unlike code for 2D and 3D
    α2 = 0.5α

    # Make sure not to include branches inside for loops.
    overwrite && (Gᵥ .= 0)
    if isfwd
        if isbloch
            # 1. At locations except for the positive end of the x-direction
            @threads for i = 1:Nx-1
                @inbounds Gᵥ[i] += (α2 * ∆w′⁻¹[i]) * (∆w[i+1]*Fᵤ[i+1] + ∆w[i]*Fᵤ[i])
            end

            # 2. At the positive end of the x-direction (where the boundary fields are taken
            # from the negative-end boundary)
            β = α2 * ∆w′⁻¹[Nx]
            @inbounds Gᵥ[Nx] += β * (∆w[1]*e⁻ⁱᵏᴸ*Fᵤ[1] + ∆w[Nx]*Fᵤ[Nx])  # Fᵤ[Nx+1] = exp(-i kx Lx) * Fᵤ[1]
        else  # symmetry boundary
            # 1. At the locations except for the positive and negative ends of the x-direction
            @threads for i = 2:Nx-1
                @inbounds Gᵥ[i] += (α2 * ∆w′⁻¹[i]) * (∆w[i+1]*Fᵤ[i+1] + ∆w[i]*Fᵤ[i])
            end

            # 2. At the negative end of the x-direction (where the boundary fields are assumed
            # zero)
            β = α2 * ∆w′⁻¹[1]
            @inbounds Gᵥ[1] += β * ∆w[2]*Fᵤ[2]  # Fᵤ[1] == 0

            # 3. At the positive end of the x-direction (where the boundary fields are assumed
            # zero)
            β = α2 * ∆w′⁻¹[Nx]
            @inbounds Gᵥ[Nx] += β * ∆w[Nx]*Fᵤ[Nx]  # Fᵤ[Nx+1] == 0
        end
    else  # backward averaging
        # 1. At the locations except for the negative end of the x-direction; unlike for the
        # forward difference, for the backward difference this part of the code is common
        # for both the Bloch and symmetry boundary conditions.
        @threads for i = 2:Nx
            @inbounds Gᵥ[i] += (α2 * ∆w′⁻¹[i]) * (∆w[i]*Fᵤ[i] + ∆w[i-1]*Fᵤ[i-1])
        end

        # 2. At the negative end of the x-direction (where the boundary fields are taken
        # from the positive-end boundary for the Bloch boundary condition)
        if isbloch
            β = α2 * ∆w′⁻¹[1]
            @inbounds Gᵥ[1] += β * (∆w[1]*Fᵤ[1] + ∆w[Nx]*Fᵤ[Nx]/e⁻ⁱᵏᴸ)  # Fᵤ[0] = Fᵤ[Nx] / exp(-i kx Lx)
        else  # symmetry boundary
            β = α * ∆w′⁻¹[1]
            @inbounds Gᵥ[1] += β * ∆w[1]*Fᵤ[1]  # Fᵤ[0] = Fᵤ[1]
        end
    end  # if isfwd

    return nothing
end
